package org.eclipse.jdt.internal.compiler.lookup;

import java.util.Collections;
import java.util.HashSet;
import java.util.Collection;
import java.util.List;
import java.util.Set;
import java.util.ArrayList;
import org.eclipse.jdt.internal.compiler.ast.ReferenceExpression;
import org.eclipse.jdt.internal.compiler.ast.LambdaExpression;
import org.eclipse.jdt.internal.compiler.ast.FunctionalExpression;

public class ConstraintExceptionFormula extends ConstraintFormula
{
    FunctionalExpression left;
    
    public ConstraintExceptionFormula(FunctionalExpression left, TypeBinding type) {
        this.left = left;
        this.right = type;
        this.relation = 7;
    }
    
    @Override
    public Object reduce(InferenceContext18 inferenceContext) {
        Scope scope = inferenceContext.scope;
        if (!this.right.isFunctionalInterface(scope)) {
            return ConstraintExceptionFormula.FALSE;
        }
        MethodBinding sam = this.right.getSingleAbstractMethod(scope, true);
        if (sam == null) {
            return ConstraintExceptionFormula.FALSE;
        }
        if ((this.left instanceof LambdaExpression)) {
            if (((LambdaExpression)this.left).argumentsTypeElided()) {
                int nParam = sam.parameters.length;
                int i = 0;
                while (i < nParam) {
                    if (!sam.parameters[i].isProperType(true)) {
                        return ConstraintExceptionFormula.FALSE;
                    }
                    ++i;
                }
            }
            if (sam.returnType != TypeBinding.VOID && !sam.returnType.isProperType(true)) {
                return ConstraintExceptionFormula.FALSE;
            }
        }
        else if (!((ReferenceExpression)this.left).isExactMethodReference()) {
            int nParam = sam.parameters.length;
            int i = 0;
            while (i < nParam) {
                if (!sam.parameters[i].isProperType(true)) {
                    return ConstraintExceptionFormula.FALSE;
                }
                ++i;
            }
            if (sam.returnType != TypeBinding.VOID && !sam.returnType.isProperType(true)) {
                return ConstraintExceptionFormula.FALSE;
            }
        }
        TypeBinding[] thrown = sam.thrownExceptions;
        InferenceVariable[] e = new InferenceVariable[thrown.length];
        int n = 0;
        for (int j = 0; j < thrown.length; ++j) {
            if (!thrown[j].isProperType(true)) {
                e[n++] = (InferenceVariable)thrown[j];
            }
        }
        if (n == 0) {
            return ConstraintExceptionFormula.TRUE;
        }
        TypeBinding[] ePrime = null;
        if ((this.left instanceof LambdaExpression)) {
            LambdaExpression lambda = ((LambdaExpression)this.left).resolveExpressionExpecting(this.right, inferenceContext.scope, inferenceContext);
            if (lambda == null) {
                return ConstraintExceptionFormula.TRUE;
            }
            Set<TypeBinding> ePrimeSet = lambda.getThrownExceptions();
            ePrime = (TypeBinding[])ePrimeSet.toArray(new TypeBinding[ePrimeSet.size()]);
        }
        else {
            ReferenceExpression referenceExpression = ((ReferenceExpression)this.left).resolveExpressionExpecting(this.right, scope, inferenceContext);
            MethodBinding method = (referenceExpression != null) ? referenceExpression.binding : null;
            if (method != null) {
                ePrime = method.thrownExceptions;
            }
        }
        if (ePrime == null) {
            return ConstraintExceptionFormula.TRUE;
        }
        int m = ePrime.length;
        List<ConstraintFormula> result = new ArrayList();
    Label_0528:
        for (int k = 0; k < m; ++k) {
            if (!ePrime[k].isUncheckedException(false)) {
                int l = 0;
                while (l < thrown.length) {
                    if (thrown[l].isProperType(true) && ePrime[k].isCompatibleWith(thrown[l])) {
                        continue Label_0528;
                    }
                    ++l;
                }
                for (l = 0; l < n; ++l) {
                    result.add(ConstraintTypeFormula.create(ePrime[k], e[l], 2));
                }
            }
        }
        for (int j2 = 0; j2 < n; ++j2) {
            inferenceContext.currentBounds.inThrows.add(e[j2].prototype());
        }
        return result.toArray(new ConstraintFormula[result.size()]);
    }
    
    @Override
    Collection<InferenceVariable> inputVariables(InferenceContext18 context) {
        if ((this.left instanceof LambdaExpression)) {
            if ((this.right instanceof InferenceVariable)) {
                return Collections.singletonList((InferenceVariable)this.right);
            }
            if (this.right.isFunctionalInterface(context.scope)) {
                LambdaExpression lambda = (LambdaExpression)this.left;
                MethodBinding sam2 = this.right.getSingleAbstractMethod(context.scope, true);
                Set<InferenceVariable> variables2 = new HashSet();
                if (lambda.argumentsTypeElided()) {
                    for (int len2 = sam2.parameters.length, j = 0; j < len2; ++j) {
                        sam2.parameters[j].collectInferenceVariables(variables2);
                    }
                }
                if (sam2.returnType != TypeBinding.VOID) {
                    sam2.returnType.collectInferenceVariables(variables2);
                    return variables2;
                }
                return variables2;
            }
        }
        else if ((this.left instanceof ReferenceExpression)) {
            if ((this.right instanceof InferenceVariable)) {
                return Collections.singletonList((InferenceVariable)this.right);
            }
            if (this.right.isFunctionalInterface(context.scope)) {
                MethodBinding sam = this.right.getSingleAbstractMethod(context.scope, true);
                Set<InferenceVariable> variables = new HashSet();
                for (int len = sam.parameters.length, i = 0; i < len; ++i) {
                    sam.parameters[i].collectInferenceVariables(variables);
                }
                sam.returnType.collectInferenceVariables(variables);
                return variables;
            }
        }
        return ConstraintExceptionFormula.EMPTY_VARIABLE_LIST;
    }
    
    @Override
    public String toString() {
        StringBuffer buf = new StringBuffer().append('⟨');
        this.left.printExpression(4, buf);
        buf.append(" ⊆throws ");
        this.appendTypeName(buf, this.right);
        buf.append('⟩');
        return buf.toString();
    }
}
